@@ -23,6 +23,8 @@
 #include "mcd-account-manager-sso.h"
 #include "mcd-debug.h"
 
+#include <telepathy-glib/util.h>
+
 #include <string.h>
 #include <ctype.h>
 
@@ -44,6 +46,7 @@
 #define PLUGIN_NAME "maemo-libaccounts"
 #define PLUGIN_DESCRIPTION \
   "Account storage in the Maemo SSO store via libaccounts-glib API"
+#define PLUGIN_PROVIDER "org.maemo.Telepathy.Account.Storage.LibAccounts"
 
 #define MCPP "param-"
 #define AGPP "parameters/"
@@ -67,6 +70,8 @@
 
 #define MC_SERVICE_KEY  "Service"
 
+static const gchar *exported_settings[] = { "CredentialsId", NULL };
+
 typedef enum {
   DELAYED_CREATE,
   DELAYED_DELETE,
@@ -186,7 +191,7 @@ static void _ag_account_stored_cb (AgAccount *acct,
     const GError *err,
     gpointer ignore);
 
-static void save_setting (AgAccount *account,
+static gboolean save_setting (AgAccount *account,
     const Setting *setting,
     const gchar *val);
 
@@ -491,9 +496,9 @@ static void _sso_toggled (GObject *object,
 
 static void _sso_deleted (GObject *object,
     AgAccountId id,
-    gpointer data)
+    gpointer user_data)
 {
-  McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (data);
+  McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (user_data);
 
   if (sso->ready)
     {
@@ -529,12 +534,18 @@ static void _sso_deleted (GObject *object,
     }
 }
 
-static void _sso_account_enable (AgAccount *account,
+/* return TRUE if we actually changed any state, FALSE otherwise */
+static gboolean _sso_account_enable (AgAccount *account,
     AgService *service,
     gboolean on)
 {
   AgService *original = ag_account_get_selected_service (account);
 
+  /* the setting account is already in one of the global+service
+     configurations that corresponds to our current state: don't touch it */
+  if (_sso_account_enabled (account, service) == on)
+    return FALSE;
+
   /* turn the local enabled flag on/off as required */
   if (service != NULL)
     ag_account_select_service (account, service);
@@ -552,6 +563,8 @@ static void _sso_account_enable (AgAccount *account,
     }
 
   ag_account_select_service (account, original);
+
+  return TRUE;
 }
 
 static gboolean _sso_account_enabled (AgAccount *account,
@@ -586,10 +599,10 @@ static gboolean _sso_account_enabled (AgAccount *account,
 
 static void _sso_created (GObject *object,
     AgAccountId id,
-    gpointer data)
+    gpointer user_data)
 {
   AgManager *ag_manager = AG_MANAGER (object);
-  McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (data);
+  McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (user_data);
   gchar *name =
     g_hash_table_lookup (sso->id_name_map, GUINT_TO_POINTER (id));
 
@@ -622,7 +635,7 @@ static void _sso_created (GObject *object,
                   g_signal_emit_by_name (mcpa, "created", name);
 
                   g_signal_connect (account, "enabled",
-                      G_CALLBACK (_sso_toggled), data);
+                      G_CALLBACK (_sso_toggled), user_data);
 
                   clear_setting_data (setting);
                 }
@@ -851,21 +864,56 @@ get_ag_account (const McdAccountManagerSso *sso,
   return NULL;
 }
 
-static void
+/* returns true if it actually changed an account's state */
+static gboolean
 save_setting (AgAccount *account,
     const Setting *setting,
     const gchar *val)
 {
+  gboolean changed = FALSE;
   AgService *service = ag_account_get_selected_service (account);
 
   if (!setting->writable)
-    return;
+    return FALSE;
 
   if (setting->global)
     ag_account_select_service (account, NULL);
   else if (service == NULL)
     _ag_account_select_default_im_service (account);
 
+  if (setting->readable)
+    {
+      GValue old = { 0 };
+      AgSettingSource src = AG_SETTING_SOURCE_NONE;
+
+      g_value_init (&old, G_TYPE_STRING);
+
+      if (setting->global)
+        src = _ag_account_global_value (account, setting->ag_name, &old);
+      else
+        src = _ag_account_local_value (account, setting->ag_name, &old);
+
+      /* unsetting an already unset value, bail out */
+      if (val == NULL && src == AG_SETTING_SOURCE_NONE)
+        goto done;
+
+      /* assigning a value to one which _is_ set: check it actually changed */
+      if (val != NULL && src != AG_SETTING_SOURCE_NONE)
+        {
+          gchar *str = _gvalue_to_string (&old);
+          gboolean noop = g_strcmp0 (str, val) == 0;
+
+          g_value_unset (&old);
+          g_free (str);
+
+          if (noop)
+            goto done;
+        }
+    }
+
+  /* if we got this far, we're changing the stored state: */
+  changed = TRUE;
+
   if (val != NULL)
     {
       GValue value = { 0 };
@@ -881,7 +929,10 @@ save_setting (AgAccount *account,
     }
 
   /* leave the selected service as we found it: */
+ done:
   ag_account_select_service (account, service);
+
+  return changed;
 }
 
 static gboolean
@@ -895,6 +946,7 @@ _set (const McpAccountStorage *self,
   McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (self);
   AgAccount *account = get_ag_account (sso, am, acct, &id);
   Setting *setting = NULL;
+  gboolean updated = FALSE;
 
   /* can't store a setting with no name */
   g_return_val_if_fail (key != NULL, FALSE);
@@ -903,7 +955,6 @@ _set (const McpAccountStorage *self,
    * in libaccount as a result of some 3rd party intervention, or it is   *
    * not an account that this plugin should ever concern itself with      */
 
-
   if (account != NULL)
     setting = setting_data (key, SETTING_MC);
   else
@@ -917,14 +968,16 @@ _set (const McpAccountStorage *self,
           gboolean on = g_str_equal (val, "true");
 
           DEBUG ("setting enabled flag: '%d'", on);
-          _sso_account_enable (account, NULL, on);
+          updated = _sso_account_enable (account, NULL, on);
         }
       else
         {
-          save_setting (account, setting, val);
+          updated = save_setting (account, setting, val);
         }
 
-      sso->save = TRUE;
+      if (updated)
+        sso->save = TRUE;
+
       clear_setting_data (setting);
     }
 
@@ -1087,6 +1140,7 @@ _delete (const McpAccountStorage *self,
   AgAccountId id;
   McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (self);
   AgAccount *account = get_ag_account (sso, am, acct, &id);
+  gboolean updated = FALSE;
 
   /* have no values for this account, nothing to do here: */
   if (account == NULL)
@@ -1100,17 +1154,21 @@ _delete (const McpAccountStorage *self,
 
       /* stop watching for updates */
       unwatch_account_keys (sso, id);
+      updated = TRUE;
     }
   else
     {
       Setting *setting = setting_data (key, SETTING_MC);
 
       if (setting != NULL)
-          save_setting (account, setting, NULL);
+        updated = save_setting (account, setting, NULL);
 
       clear_setting_data (setting);
     }
 
+  if (updated)
+    sso->save = TRUE;
+
   return TRUE;
 }
 
@@ -1238,6 +1296,9 @@ _load_from_libaccounts (McdAccountManagerSso *sso,
               mcp_account_manager_set_value (am, name, MC_PROTOCOL_KEY, mc_id[1]);
               mcp_account_manager_set_value (am, name, MC_IDENTITY_KEY, name);
 
+              /* force the services value to be synthesised + cached */
+              _get (MCP_ACCOUNT_STORAGE (sso), am, name, SERVICES_KEY);
+
               ag_account_select_service (account, service);
 
               g_signal_connect (account, "enabled",
@@ -1339,6 +1400,115 @@ _ready (const McpAccountStorage *self,
   sso->pending_signals = NULL;
 }
 
+static gboolean
+_find_account (McdAccountManagerSso *sso,
+    const gchar *account_name,
+    AgAccountId *account_id)
+{
+  GList *ag_ids = NULL;
+  GList *ag_id;
+  gboolean found = FALSE;
+
+  g_return_val_if_fail (account_id != NULL, found);
+
+  ag_ids = ag_manager_list_by_service_type (sso->ag_manager, "IM");
+
+  for (ag_id = ag_ids; ag_id != NULL; ag_id = g_list_next (ag_id))
+    {
+      AgAccountId id = GPOINTER_TO_UINT (ag_id->data);
+      gchar *name = NULL;
+
+      name = _ag_accountid_to_mc_key (sso, id, FALSE);
+
+      if (g_strcmp0 (name, account_name) == 0)
+        {
+          found = TRUE;
+          *account_id = id;
+        }
+
+      g_free (name);
+
+      if (found)
+        break;
+    }
+
+  ag_manager_list_free (ag_ids);
+
+  return found;
+}
+
+static void
+_get_identifier (const McpAccountStorage *self,
+    const gchar *account,
+    GValue *identifier)
+{
+  AgAccountId account_id = 0;
+
+  if (!_find_account (MCD_ACCOUNT_MANAGER_SSO (self), account, &account_id))
+    g_warning ("Didn't find account %s in %s", account, PLUGIN_NAME);
+
+  g_value_init (identifier, G_TYPE_UINT);
+
+  g_value_set_uint (identifier, account_id);
+}
+
+static GHashTable *
+_get_additional_info (const McpAccountStorage *self,
+    const gchar *account)
+{
+  AgAccountId account_id = 0;
+  McdAccountManagerSso *sso = MCD_ACCOUNT_MANAGER_SSO (self);
+  GHashTable *additional_info = NULL;
+  AgAccount *acct;
+  AgService *service;
+  AgAccountSettingIter iter;
+  const GValue *val;
+  const gchar *key;
+
+  if (!_find_account (sso, account, &account_id))
+    {
+      g_warning ("Didn't find account %s in %s", account, PLUGIN_NAME);
+      return NULL;
+    }
+
+  acct = ag_manager_get_account (sso->ag_manager, account_id);
+
+  g_return_val_if_fail (acct != NULL, NULL);
+
+  service = ag_account_get_selected_service (acct);
+
+  additional_info = g_hash_table_new_full (g_str_hash, g_str_equal,
+      (GDestroyNotify) g_free, (GDestroyNotify) tp_g_value_slice_free);
+
+  if (service == NULL)
+    _ag_account_select_default_im_service (acct);
+
+  ag_account_settings_iter_init (acct, &iter, NULL);
+
+  while (ag_account_settings_iter_next (&iter, &key, &val))
+    {
+      if (tp_strv_contains (exported_settings, key))
+          g_hash_table_insert (additional_info, g_strdup (key),
+              tp_g_value_slice_dup (val));
+    }
+
+  ag_account_select_service (acct, NULL);
+  ag_account_settings_iter_init (acct, &iter, NULL);
+
+  while (ag_account_settings_iter_next (&iter, &key, &val))
+    {
+      if (tp_strv_contains (exported_settings, key))
+          g_hash_table_insert (additional_info, g_strdup (key),
+              tp_g_value_slice_dup (val));
+    }
+
+  ag_account_select_service (acct, service);
+
+  g_object_unref (acct);
+
+  return additional_info;
+}
+
 static void
 account_storage_iface_init (McpAccountStorageIface *iface,
     gpointer unused G_GNUC_UNUSED)
@@ -1346,6 +1516,7 @@ account_storage_iface_init (McpAccountStorageIface *iface,
   mcp_account_storage_iface_set_name (iface, PLUGIN_NAME);
   mcp_account_storage_iface_set_desc (iface, PLUGIN_DESCRIPTION);
   mcp_account_storage_iface_set_priority (iface, PLUGIN_PRIORITY);
+  mcp_account_storage_iface_set_provider (iface, PLUGIN_PROVIDER);
 
   mcp_account_storage_iface_implement_get (iface, _get);
   mcp_account_storage_iface_implement_set (iface, _set);
@@ -1353,6 +1524,9 @@ account_storage_iface_init (McpAccountStorageIface *iface,
   mcp_account_storage_iface_implement_commit (iface, _commit);
   mcp_account_storage_iface_implement_list (iface, _list);
   mcp_account_storage_iface_implement_ready (iface, _ready);
+  mcp_account_storage_iface_implement_get_identifier (iface, _get_identifier);
+  mcp_account_storage_iface_implement_get_additional_info (iface,
+      _get_additional_info);
 }
 
 McdAccountManagerSso *
