@@ -3,8 +3,8 @@
 /*
  * This file is part of mission-control
  *
- * Copyright (C) 2007-2009 Nokia Corporation.
- * Copyright (C) 2009 Collabora Ltd.
+ * Copyright © 2007-2009 Nokia Corporation.
+ * Copyright © 2009-2010 Collabora Ltd.
  *
  * Contact: Naba Kumar  <naba.kumar@nokia.com>
  *
@@ -75,7 +75,9 @@
 #include "_gen/cli-Connection_Interface_Contact_Capabilities_Draft1-body.h"
 #include "_gen/cli-Connection_Interface_Contact_Capabilities-body.h"
 
-#define INITIAL_RECONNECTION_TIME   1 /* 1 second */
+#define INITIAL_RECONNECTION_TIME   3 /* seconds */
+#define RECONNECTION_MULTIPLIER     3
+#define MAXIMUM_RECONNECTION_TIME   30 * 60 /* half an hour */
 
 #define MCD_CONNECTION_PRIV(mcdconn) (MCD_CONNECTION (mcdconn)->priv)
 
@@ -138,6 +140,9 @@ struct _McdConnectionPrivate
     /* FALSE until mcd_connection_close() is called */
     guint closed : 1;
 
+    /* FALSE until connected and the supported presence statuses retrieved */
+    guint presence_info_ready : 1;
+
     gchar *alias;
 
     gboolean is_disposed;
@@ -200,6 +205,8 @@ presence_set_status_cb (TpConnection *proxy, const GError *error,
 
     if (error)
     {
+        _mcd_account_set_changing_presence (priv->account, FALSE);
+
         g_warning ("%s: Setting presence of %s failed: %s",
 		   G_STRFUNC, mcd_account_get_unique_name (priv->account),
                    error->message);
@@ -311,9 +318,24 @@ _mcd_connection_set_presence (McdConnection * connection,
 
     if (_check_presence (priv, presence, &adj_status))
     {
+        TpConnectionPresenceType curr_presence;
+        const gchar *curr_status;
+        const gchar *curr_message;
+
         DEBUG ("Setting status '%s' of type %u ('%s' was requested)",
                adj_status, presence, status);
 
+        mcd_account_get_current_presence (priv->account, &curr_presence,
+                                          &curr_status, &curr_message);
+        if (curr_presence == presence &&
+            tp_strdiff (curr_status, adj_status) == 0 &&
+            tp_strdiff (curr_message, message) == 0)
+        {
+            // PresencesChanged won't be emitted and Account.ChangingPresence
+            // will never go to FALSE, so forcibly set it to FALSE
+            _mcd_account_set_changing_presence (priv->account, FALSE);
+        }
+
         tp_cli_connection_interface_simple_presence_call_set_presence
             (priv->tp_conn, -1, adj_status, message, presence_set_status_cb,
              priv, NULL, (GObject *)connection);
@@ -389,8 +411,13 @@ presence_get_statuses_cb (TpProxy *proxy, const GValue *v_statuses,
     }
 
     /* Now the presence info is ready. We can set the presence */
-    mcd_account_get_requested_presence (priv->account, &presence,
+    _mcd_account_get_combined_presence (priv->account, &presence,
                                         &status, &message);
+    if (priv->connected)
+    {
+        priv->presence_info_ready = TRUE;
+    }
+
     _mcd_connection_set_presence (connection, presence, status, message);
 }
 
@@ -672,6 +699,7 @@ avatars_clear_avatar_cb (TpConnection *proxy, const GError *error,
     }
 }
 
+/* this signal handler only deals with our own avatar updates */
 static void
 on_avatar_retrieved (TpConnection *proxy, guint contact_id, const gchar *token,
 		     const GArray *avatar, const gchar *mime_type,
@@ -707,6 +735,54 @@ avatars_request_avatars_cb (TpConnection *proxy, const GError *error,
     }
 }
 
+static void
+avatars_known_token_cb (TpConnection *proxy, GHashTable *tokens,
+                        const GError *error, gpointer user_data,
+                        GObject *weak_object)
+{
+    McdConnection *connection = MCD_CONNECTION (weak_object);
+    McdConnectionPrivate *priv = connection->priv;
+    const gchar *token;
+    TpHandle self_handle = tp_connection_get_self_handle (proxy);
+    gpointer handle = user_data;
+
+    if (error)
+    {
+        g_warning ("%s: error: %s", G_STRFUNC, error->message);
+        return;
+    }
+
+    if (handle != GUINT_TO_POINTER (self_handle))
+        return;
+
+    token = g_hash_table_lookup (tokens, handle);
+
+    /* we have a token and it is not the empty string (ie no avatar) */
+    if (token != NULL && *token != '\0')
+    {
+        GArray handles = { (gchar *) &handle, 1 };
+
+        /* if we have an avatar, set off the request to fetch it: this will *
+           get picked up as we connect to the retrieved signal elsewhere    */
+        tp_cli_connection_interface_avatars_call_request_avatars (priv->tp_conn,
+                                                                  -1,
+                                                                  &handles,
+                                                                  avatars_request_avatars_cb,
+                                                                  NULL,
+                                                                  NULL,
+                                                                  weak_object);
+    }
+    else
+    {   /* no avatar => will never get a retrieved signal - remove manually */
+        GError *avatar_error = NULL;
+        McdAccount *account = mcd_connection_get_account (connection);
+
+        if (!_mcd_account_set_avatar (account, NULL, "", "", &avatar_error))
+            DEBUG ("Attempt to clear avatar failed: %s", avatar_error->message);
+    }
+}
+
+/* this signal handler only deals with our own avatar updates */
 static void
 on_avatar_updated (TpConnection *proxy, guint contact_id, const gchar *token,
 		   gpointer user_data, GObject *weak_object)
@@ -725,16 +801,16 @@ on_avatar_updated (TpConnection *proxy, guint contact_id, const gchar *token,
 
     if (!prev_token || strcmp (token, prev_token) != 0)
     {
-    	GArray handles;
-        DEBUG ("avatar has changed");
-	/* the avatar has changed, let's retrieve the new one */
-	handles.len = 1;
-	handles.data = (gchar *)&contact_id;
-	tp_cli_connection_interface_avatars_call_request_avatars (priv->tp_conn, -1,
-								  &handles,
-								  avatars_request_avatars_cb,
-								  priv, NULL,
-								  (GObject *)connection);
+        GArray handles = { (gchar *) &contact_id, 1 };
+
+        DEBUG ("avatar has changed or been erased");
+        tp_cli_connection_interface_avatars_call_get_known_avatar_tokens (priv->tp_conn,
+                                                                          -1,
+                                                                          &handles,
+                                                                          avatars_known_token_cb,
+                                                                          GUINT_TO_POINTER (contact_id),
+                                                                          NULL,
+                                                                          (GObject *)connection);
     }
     g_free (prev_token);
 }
@@ -1002,15 +1078,26 @@ mcd_connection_reconnect (McdConnection *connection)
 static gboolean
 mcd_connection_probation_ended_cb (gpointer user_data)
 {
-    McdConnection *self = user_data;
+    McdConnection *self = MCD_CONNECTION (user_data);
+    McdConnectionPrivate *priv = MCD_CONNECTION_PRIV (self);
 
     /* We've been connected for PROBATION_SEC seconds. We can probably now
      * assume that the connection is stable */
-    DEBUG ("probation finished, assuming connection is stable: %s",
-           tp_proxy_get_object_path (self->priv->tp_conn));
+    if (priv->tp_conn != NULL)
+    {
+        DEBUG ("probation finished, assuming connection is stable: %s",
+               tp_proxy_get_object_path (self->priv->tp_conn));
+        self->priv->probation_drop_count = 0;
+        self->priv->reconnect_interval = INITIAL_RECONNECTION_TIME;
+    }
+    else /* probation timer survived beyond its useful life */
+    {
+        g_warning ("probation error: timer should have been removed when the "
+                   "TpConnection was released");
+    }
+
     self->priv->probation_timer = 0;
-    self->priv->probation_drop_count = 0;
-    self->priv->reconnect_interval = INITIAL_RECONNECTION_TIME;
+
     return FALSE;
 }
 
@@ -1032,7 +1119,7 @@ on_connection_status_changed (TpConnection *tp_conn, GParamSpec *pspec,
     {
     case TP_CONNECTION_STATUS_CONNECTING:
         g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
-                       conn_status, conn_reason);
+                       conn_status, conn_reason, tp_conn);
         priv->abort_reason = TP_CONNECTION_STATUS_REASON_NONE_SPECIFIED;
         priv->connected = FALSE;
         break;
@@ -1040,7 +1127,7 @@ on_connection_status_changed (TpConnection *tp_conn, GParamSpec *pspec,
     case TP_CONNECTION_STATUS_CONNECTED:
         {
             g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
-                           conn_status, conn_reason);
+                           conn_status, conn_reason, tp_conn);
 
             if (priv->probation_timer == 0)
             {
@@ -1110,14 +1197,15 @@ mcd_connection_invalidated_cb (TpConnection *tp_conn,
          * abort the connection but try to reconnect later */
         if (priv->reconnect_timer == 0)
         {
-            DEBUG ("Preparing for reconnection");
+            DEBUG ("Preparing for reconnection in %u seconds",
+                priv->reconnect_interval);
             priv->reconnect_timer = g_timeout_add_seconds
                 (priv->reconnect_interval,
                  (GSourceFunc)mcd_connection_reconnect, connection);
-            priv->reconnect_interval *= 2;
-            if (priv->reconnect_interval >= 30 * 60)
-                /* no more than 30 minutes! */
-                priv->reconnect_interval = 30 * 60;
+            priv->reconnect_interval *= RECONNECTION_MULTIPLIER;
+
+            if (priv->reconnect_interval >= MAXIMUM_RECONNECTION_TIME)
+                priv->reconnect_interval = MAXIMUM_RECONNECTION_TIME;
         }
     }
     else
@@ -1168,22 +1256,22 @@ request_unrequested_channels (McdConnection *connection)
 }
 
 static McdChannel *
-find_channel_by_path (McdConnection *connection, const gchar *object_path)
+find_channel_by_path (McdConnection *connection,
+                      const gchar *object_path)
 {
     const GList *list = NULL;
 
     list = mcd_operation_get_missions (MCD_OPERATION (connection));
+
     while (list)
     {
         McdChannel *channel = MCD_CHANNEL (list->data);
-        const gchar *req_object_path;
 
-        req_object_path = mcd_channel_get_object_path (channel);
-        if (req_object_path &&
-            strcmp (object_path, req_object_path) == 0)
+        if (_mcd_channel_is_primary_for_path (channel, object_path))
         {
             return channel;
         }
+
         list = list->next;
     }
     return NULL;
@@ -1532,7 +1620,10 @@ on_connection_ready (TpConnection *tp_conn, const GError *error,
     if (priv->has_alias_if)
 	_mcd_connection_setup_alias (connection);
 
-    _mcd_dispatcher_add_connection (priv->dispatcher, connection);
+    if (!priv->dispatching_started)
+        _mcd_dispatcher_add_connection (priv->dispatcher, connection);
+
+    request_unrequested_channels (connection);
 
     g_signal_emit (connection, signals[READY], 0);
 }
@@ -1554,10 +1645,9 @@ _mcd_connection_start_dispatching (McdConnection *self,
     else
         mcd_connection_setup_pre_requests (self);
 
+    /* FIXME: why is this here? if we need to update caps before and after   *
+     * connected, it should be in the call_when_ready callback.              */
     _mcd_connection_update_client_caps (self, client_caps);
-
-    /* and request all channels */
-    request_unrequested_channels (self);
 }
 
 void
@@ -1587,6 +1677,12 @@ mcd_connection_done_task_before_connect (McdConnection *self)
             DEBUG ("TpConnection went away, not doing anything");
         }
 
+        if (tp_proxy_has_interface_by_id (self->priv->tp_conn,
+                TP_IFACE_QUARK_CONNECTION_INTERFACE_REQUESTS))
+        {
+            _mcd_dispatcher_add_connection (self->priv->dispatcher, self);
+        }
+
         DEBUG ("%s: Calling Connect()",
                tp_proxy_get_object_path (self->priv->tp_conn));
         tp_cli_connection_call_connect (self->priv->tp_conn, -1, connect_cb,
@@ -1706,6 +1802,12 @@ mcd_connection_early_get_interfaces_cb (TpConnection *tp_conn,
                  * we can't usefully pre-load capabilities - we'll be told
                  * the real capabilities as soon as it has worked them out */
             }
+            else if (q == TP_IFACE_QUARK_CONNECTION_INTERFACE_REQUESTS)
+            {
+              /* If we have the Requests iface, we could start dispatching
+               * before the connection is in CONNECTED state */
+              tp_proxy_add_interface_by_id ((TpProxy *) tp_conn, q);
+            }
         }
     }
 
@@ -1781,7 +1883,7 @@ request_connection_cb (TpConnectionManager *proxy, const gchar *bus_name,
         {
             g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
                            TP_CONNECTION_STATUS_DISCONNECTED,
-                           TP_CONNECTION_STATUS_REASON_REQUESTED);
+                           TP_CONNECTION_STATUS_REASON_REQUESTED, NULL);
         }
 
         return;
@@ -1796,7 +1898,7 @@ request_connection_cb (TpConnectionManager *proxy, const gchar *bus_name,
 
         g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
             TP_CONNECTION_STATUS_DISCONNECTED,
-            TP_CONNECTION_STATUS_REASON_NETWORK_ERROR);
+            TP_CONNECTION_STATUS_REASON_NETWORK_ERROR, NULL);
         return;
     }
 
@@ -1833,7 +1935,7 @@ _mcd_connection_connect_with_params (McdConnection *connection,
 
     g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
                    TP_CONNECTION_STATUS_CONNECTING,
-                   TP_CONNECTION_STATUS_REASON_REQUESTED);
+                   TP_CONNECTION_STATUS_REASON_REQUESTED, NULL);
 
     /* If the McdConnection gets aborted (which results in it being freed!),
      * we need to kill off the Connection. So, we can't use connection as the
@@ -1869,9 +1971,11 @@ _mcd_connection_release_tp_connection (McdConnection *connection)
     DEBUG ("%p", connection);
     g_signal_emit (connection, signals[SELF_PRESENCE_CHANGED], 0,
                    TP_CONNECTION_PRESENCE_TYPE_OFFLINE, "offline", "");
+
     g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
                    TP_CONNECTION_STATUS_DISCONNECTED,
-                   priv->abort_reason);
+                   priv->abort_reason, priv->tp_conn);
+
     if (priv->tp_conn)
     {
 	/* Disconnect signals */
@@ -1882,9 +1986,16 @@ _mcd_connection_release_tp_connection (McdConnection *connection)
             G_CALLBACK (mcd_connection_invalidated_cb), connection);
 
 	_mcd_connection_call_disconnect (connection);
-	g_object_unref (priv->tp_conn);
-	priv->tp_conn = NULL;
-	_mcd_account_tp_connection_changed (priv->account);
+
+        /* the tp_connection has gone away, so we no longer need (or want) *
+           the probation timer to go off: there's nothing for it to check  */
+        if (priv->probation_timer > 0)
+        {
+            g_source_remove (priv->probation_timer);
+            priv->probation_timer = 0;
+        }
+
+        tp_clear_object (&priv->tp_conn);
     }
 
     /* the interface proxies obtained from this connection must be deleted, too
@@ -1944,27 +2055,12 @@ _mcd_connection_dispose (GObject * object)
         g_signal_handlers_disconnect_by_func (priv->account,
                                               G_CALLBACK (on_account_removed),
                                               object);
-	g_object_unref (priv->account);
-	priv->account = NULL;
+        tp_clear_object (&priv->account);
     }
 
-    if (priv->tp_conn_mgr)
-    {
-	g_object_unref (priv->tp_conn_mgr);
-	priv->tp_conn_mgr = NULL;
-    }
-
-    if (priv->dispatcher)
-    {
-	g_object_unref (priv->dispatcher);
-	priv->dispatcher = NULL;
-    }
-
-    if (priv->dbus_daemon)
-    {
-	g_object_unref (priv->dbus_daemon);
-	priv->dbus_daemon = NULL;
-    }
+    tp_clear_object (&priv->tp_conn_mgr);
+    tp_clear_object (&priv->dispatcher);
+    tp_clear_object (&priv->dbus_daemon);
 
     G_OBJECT_CLASS (mcd_connection_parent_class)->dispose (object);
 }
@@ -1987,22 +2083,17 @@ _mcd_connection_set_property (GObject * obj, guint prop_id,
 	    g_return_if_fail (MCD_IS_DISPATCHER (dispatcher));
 	    g_object_ref (dispatcher);
 	}
-	if (priv->dispatcher)
-	{
-	    g_object_unref (priv->dispatcher);
-	}
+	tp_clear_object (&priv->dispatcher);
 	priv->dispatcher = dispatcher;
 	break;
     case PROP_DBUS_DAEMON:
-	if (priv->dbus_daemon)
-	    g_object_unref (priv->dbus_daemon);
+	tp_clear_object (&priv->dbus_daemon);
 	priv->dbus_daemon = TP_DBUS_DAEMON (g_value_dup_object (val));
 	break;
     case PROP_TP_MANAGER:
 	tp_conn_mgr = g_value_get_object (val);
 	g_object_ref (tp_conn_mgr);
-	if (priv->tp_conn_mgr)
-	    g_object_unref (priv->tp_conn_mgr);
+	tp_clear_object (&priv->tp_conn_mgr);
 	priv->tp_conn_mgr = tp_conn_mgr;
 	break;
     case PROP_ACCOUNT:
@@ -2162,7 +2253,7 @@ mcd_connection_class_init (McdConnectionClass * klass)
     klass->need_dispatch = mcd_connection_need_dispatch;
     klass->request_channel = _mcd_connection_request_channel;
 
-    _mc_ext_register_dbus_glib_marshallers ();
+    _mcd_ext_register_dbus_glib_marshallers ();
 
     tp_connection_init_known_interfaces ();
     tp_proxy_or_subclass_hook_on_interface_add (TP_TYPE_CONNECTION,
@@ -2218,8 +2309,8 @@ mcd_connection_class_init (McdConnectionClass * klass)
     signals[CONNECTION_STATUS_CHANGED] = g_signal_new (
         "connection-status-changed", G_OBJECT_CLASS_TYPE (klass),
         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED, 0,
-        NULL, NULL, _mcd_marshal_VOID__UINT_UINT,
-        G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_UINT);
+        NULL, NULL, _mcd_marshal_VOID__UINT_UINT_OBJECT,
+        G_TYPE_NONE, 3, G_TYPE_UINT, G_TYPE_UINT, TP_TYPE_CONNECTION);
 
     signals[READY] = g_signal_new ("ready",
         G_OBJECT_CLASS_TYPE (klass),
@@ -2387,6 +2478,17 @@ mcd_connection_request_channel (McdConnection *connection,
     g_return_val_if_fail (MCD_IS_CONNECTION (connection), FALSE);
     g_return_val_if_fail (MCD_IS_CHANNEL (channel), FALSE);
 
+    if (mcd_channel_get_status (channel) == MCD_CHANNEL_STATUS_FAILED)
+    {
+        DEBUG ("Channel %p failed already, never mind", channel);
+        _mcd_channel_close (channel);
+        mcd_mission_abort (MCD_MISSION (channel));
+        /* FIXME: the boolean return is a decoy - everyone returns TRUE and
+         * every caller ignores it - so it's not clear what FALSE would
+         * mean. */
+        return TRUE;
+    }
+
     if (!mcd_mission_get_parent ((McdMission *)channel))
         mcd_operation_take_mission (MCD_OPERATION (connection),
                                     MCD_MISSION (channel));
@@ -2572,10 +2674,13 @@ _mcd_connection_set_tp_connection (McdConnection *connection,
     {
         g_signal_emit (connection, signals[CONNECTION_STATUS_CHANGED], 0,
             TP_CONNECTION_STATUS_DISCONNECTED,
-            TP_CONNECTION_STATUS_REASON_NETWORK_ERROR);
+            TP_CONNECTION_STATUS_REASON_NETWORK_ERROR,
+            NULL);
         return;
     }
-    _mcd_account_tp_connection_changed (priv->account);
+    /* FIXME: need some way to feed the status into the Account, but we don't
+     * actually know it yet */
+    _mcd_account_tp_connection_changed (priv->account, priv->tp_conn);
 
     /* Setup signals */
     g_signal_connect (priv->tp_conn, "invalidated",
@@ -2614,3 +2719,11 @@ _mcd_connection_is_ready (McdConnection *self)
     return (self->priv->tp_conn != NULL) &&
         tp_connection_is_ready (self->priv->tp_conn);
 }
+
+gboolean
+_mcd_connection_presence_info_is_ready (McdConnection *self)
+{
+    g_return_val_if_fail (MCD_IS_CONNECTION (self), FALSE);
+
+    return self->priv->presence_info_ready;
+}
