@@ -45,9 +45,7 @@
 #include <string.h>
 
 #include <glib/gi18n.h>
-#include <telepathy-glib/connection-manager.h>
-#include <telepathy-glib/dbus.h>
-#include <telepathy-glib/interfaces.h>
+#include <telepathy-glib/telepathy-glib.h>
 
 #include <libmcclient/mc-errors.h>
 
@@ -147,20 +145,9 @@ _mcd_manager_dispose (GObject * object)
 
     priv->is_disposed = TRUE;
 
-    if (priv->dispatcher)
-    {
-	g_object_unref (priv->dispatcher);
-	priv->dispatcher = NULL;
-    }
-    
-    if (priv->tp_conn_mgr)
-    {
-	g_object_unref (priv->tp_conn_mgr);
-	priv->tp_conn_mgr = NULL;
-    }
-
-    if (priv->dbus_daemon)
-	g_object_unref (priv->dbus_daemon);
+    tp_clear_object (&priv->dispatcher);
+    tp_clear_object (&priv->tp_conn_mgr);
+    tp_clear_object (&priv->dbus_daemon);
 
     G_OBJECT_CLASS (mcd_manager_parent_class)->dispose (object);
 }
@@ -216,10 +203,8 @@ mcd_manager_setup (McdManager *manager)
     return TRUE;
 
 error:
-    if (priv->tp_conn_mgr)
-        g_object_unref (priv->tp_conn_mgr);
-    if (error)
-        g_error_free (error);
+    tp_clear_object (&priv->tp_conn_mgr);
+    g_clear_error (&error);
 
     return FALSE;
 }
@@ -266,15 +251,11 @@ _mcd_manager_set_property (GObject * obj, guint prop_id,
 	    g_return_if_fail (MCD_IS_DISPATCHER (dispatcher));
 	    g_object_ref (dispatcher);
 	}
-	if (priv->dispatcher)
-	{
-	    g_object_unref (priv->dispatcher);
-	}
+	tp_clear_object (&priv->dispatcher);
 	priv->dispatcher = dispatcher;
 	break;
     case PROP_DBUS_DAEMON:
-	if (priv->dbus_daemon)
-	    g_object_unref (priv->dbus_daemon);
+	tp_clear_object (&priv->dbus_daemon);
 	priv->dbus_daemon = TP_DBUS_DAEMON (g_value_dup_object (val));
 	break;
     default:
@@ -474,77 +455,6 @@ mcd_manager_get_parameters (McdManager *manager, const gchar *protocol)
     return NULL;
 }
 
-/* Copied from telepathy-glib 0.11.6 for _mcd_manager_copy_protocol() */
-static void
-_tp_connection_manager_param_free_contents (TpConnectionManagerParam *param)
-{
-  g_free (param->name);
-  g_free (param->dbus_signature);
-
-  if (G_IS_VALUE (&param->default_value))
-    g_value_unset (&param->default_value);
-}
-
-/* Copied from telepathy-glib 0.11.6 for _mcd_manager_copy_protocol() */
-void
-_mcd_manager_protocol_free (TpConnectionManagerProtocol *proto)
-{
-  TpConnectionManagerParam *param;
-
-  g_free (proto->name);
-
-  for (param = proto->params; param->name != NULL; param++)
-    {
-      _tp_connection_manager_param_free_contents (param);
-    }
-
-  g_free (proto->params);
-
-  g_slice_free (TpConnectionManagerProtocol, proto);
-}
-
-/* Copied from telepathy-glib 0.11.6 for _mcd_manager_copy_protocol() */
-static void
-_tp_connection_manager_param_copy_contents (
-    const TpConnectionManagerParam *in,
-    TpConnectionManagerParam *out)
-{
-  out->name = g_strdup (in->name);
-  out->dbus_signature = g_strdup (in->dbus_signature);
-  out->flags = in->flags;
-
-  if (G_IS_VALUE (&in->default_value))
-    {
-      g_value_init (&out->default_value, G_VALUE_TYPE (&in->default_value));
-      g_value_copy (&in->default_value, &out->default_value);
-    }
-}
-
-/* Copied from telepathy-glib 0.11.6's tp_connection_manager_protocol_copy();
- * duplicated here to avoid a newer dependency in the stable branch. */
-static TpConnectionManagerProtocol *
-_mcd_manager_copy_protocol (const TpConnectionManagerProtocol *in)
-{
-  TpConnectionManagerProtocol *out = g_slice_new0 (TpConnectionManagerProtocol);
-  TpConnectionManagerParam *param;
-  GArray *params = g_array_new (TRUE, TRUE,
-      sizeof (TpConnectionManagerParam));
-
-  out->name = g_strdup (in->name);
-
-  for (param = in->params; param->name != NULL; param++)
-    {
-      TpConnectionManagerParam copy = { 0, };
-
-      _tp_connection_manager_param_copy_contents (param, &copy);
-      g_array_append_val (params, copy);
-    }
-
-  out->params = (TpConnectionManagerParam *) g_array_free (params, FALSE);
-
-  return out;
-}
-
 TpConnectionManagerProtocol *
 _mcd_manager_dup_protocol (McdManager *manager,
                            const gchar *protocol)
@@ -559,7 +469,7 @@ _mcd_manager_dup_protocol (McdManager *manager,
     if (p == NULL)
         return NULL;
     else
-        return _mcd_manager_copy_protocol (p);
+        return tp_connection_manager_protocol_copy (p);
 }
 
 const TpConnectionManagerParam *
